import requests
import logging
from typing import Optional

logger = logging.getLogger(__name__)


class TokenManager:
    """Manages OAuth tokens for Netatmo API with automatic refresh"""

    def __init__(self, config):
        self.config = config
        self.cfg = config['config']

    def get_valid_token(self) -> Optional[str]:
        """Get access token from configuration (no validation)"""
        access_token = self.cfg.get('access_token')

        if not access_token:
            logger.error("No access_token found in configuration")
            return None

        return access_token

    def refresh_on_auth_error(self) -> bool:
        """Refresh token after authentication error"""
        logger.info("Authentication failed, attempting token refresh")
        return self._refresh_access_token()

    def _refresh_access_token(self) -> bool:
        """Refresh access token using refresh_token"""
        refresh_token = self.cfg.get('refresh_token')

        if not refresh_token:
            logger.error("No refresh_token found in configuration")
            return False

        payload = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
            'client_id': self.cfg['client_id'],
            'client_secret': self.cfg['client_secret'],
        }

        try:
            response = requests.post("https://api.netatmo.com/oauth2/token", data=payload, timeout=10)
            response.raise_for_status()

            data = response.json()

            # Check if tokens actually changed
            access_token_changed = data['access_token'] != self.cfg.get('access_token')
            refresh_token_changed = 'refresh_token' in data and data['refresh_token'] != self.cfg.get('refresh_token')

            # Update configuration only if tokens changed
            if access_token_changed or refresh_token_changed:
                self.cfg['access_token'] = data['access_token']
                if 'refresh_token' in data:
                    self.cfg['refresh_token'] = data['refresh_token']

                # Save updated configuration only when needed
                self.config.write()

                if access_token_changed:
                    logger.info("Access token refreshed")
                if refresh_token_changed:
                    logger.info("Refresh token updated")
            else:
                logger.info("Tokens unchanged (no refresh needed)")

            return True

        except requests.exceptions.RequestException as e:
            logger.error(f"Token refresh failed: {e}")
            return False
        except KeyError as e:
            logger.error(f"Invalid token refresh response: {e}")
            return False



from xaal.lib import tools, Device, DeviceError

import requests
import logging

from .token_manager import TokenManager

logger = logging.getLogger(__name__)


class API:
    """NetAtmo API"""

    def __init__(self, configfile):
        self.configfile = configfile
        self.token_manager = TokenManager(configfile)

    def get_token(self):
        """Get valid access token using TokenManager"""
        return self.token_manager.get_valid_token()

    def debug_enabled(self):
        if self.config().get('debug', '') == "True":
            return True
        return False

    def config(self):
        return self.configfile['config']

    def get_data(self):  # Fonction which get the information from netatmo API by using a token
        token = self.get_token()
        if token is None:
            return None

        params = {'access_token': token}
        try:
            response = requests.post("https://api.netatmo.com/api/getstationsdata", params=params)

            # Handle authentication errors
            if response.status_code in [401, 403]:
                if self.token_manager.refresh_on_auth_error():
                    # Retry with refreshed token
                    token = self.get_token()
                    if token:
                        params['access_token'] = token
                        response = requests.post("https://api.netatmo.com/api/getstationsdata", params=params)

            response.raise_for_status()
            data = response.json()["body"]

            # Debug dump if enabled
            if self.debug_enabled():
                logger.info("=== NETATMO DATA DUMP ===")
                import json

                logger.info(json.dumps(data, indent=2))
                logger.info("=== END DUMP ===")

            return data
        except requests.exceptions.RequestException as e:
            logger.error(f"API request failed: {e}")
            return None


class NAModule(object):  # NWS: netatmo weather station
    def __init__(self, base_addr, group, macId=None, na_type=None):
        self.sensors = []
        self.base_addr = base_addr
        self.group = group
        self.macId = macId
        self.na_type = na_type

    def build_dev(self, dtype, addr):
        dev = Device(dtype)
        dev.dev_type = dtype
        dev.address = addr
        dev.vendor_id = "NETATMO"
        dev.product_id = "Netatmo:%s" % self.na_type
        dev.info = "mac:%s" % self.macId
        return dev

    def new_sensor(self, sensor_type, sensor_attr, senor_addr, sensor_group):
        s_type = sensor_type + ".basic"
        sensor = self.build_dev(s_type, senor_addr)
        for attr in sensor_attr:
            sensor.new_attribute(attr)
        sensor.group_id = sensor_group
        return sensor

    def update_sensor(self, sensor, sensor_attr, attr_value):
        if isinstance(sensor, Device):
            attr = sensor.get_attribute(sensor_attr)
            if attr:
                attr.value = attr_value
        else:
            raise DeviceError("This sensor is invalid")

    def new_sensors(self, sensors_type_list):
        for i in range(len(sensors_type_list)):
            sensor_type = sensors_type_list[i]
            sensor_attr = self.get_sensor_attr(sensor_type)
            sensor_addr = self.base_addr + i + 1
            sensor = self.new_sensor(sensor_type, sensor_attr, sensor_addr, self.group)
            self.sensors.append(sensor)

    def update_sensors(self, attr_value_dic):
        for sensor in self.sensors:
            for sensor_attr, attr_value in attr_value_dic.items():
                self.update_sensor(sensor, sensor_attr, attr_value)

    def get_sensor_attr(self, sensor_type):
        if sensor_type == "thermometer":
            return ["temperature"]
        elif sensor_type == "hygrometer":
            return ["humidity"]
        elif sensor_type == "barometer":
            return ["pressure"]
        elif sensor_type == "co2meter":
            return ["co2"]
        elif sensor_type == "soundmeter":
            return ["sound"]
        elif sensor_type == "windgauge":
            return ["windstrength", "windangle", "guststrength", "gustangle"]
        elif sensor_type == "rainmeter":
            return ["rain"]
        elif sensor_type == "radiometer":
            return ["radio_status"]
        elif sensor_type == "wifimeter":
            return ["rssi"]
        elif sensor_type == "battery":
            return ["level"]
        else:
            logger.warning("invalid sensor type !")


class ConfigParser:  # netatmo weather station config
    def __init__(self, configfile):
        self.cfg = configfile

    def update(self):
        if 'config' not in self.cfg:
            logger.error("invalid config file !!!")
        if 'addr' not in self.cfg['config']:
            self.cfg['config']['addr'] = tools.get_random_uuid()
        if 'devices' not in self.cfg:
            self.cfg['devices'] = {}

        self.update_devices()
        self.cfg.write()

    def update_devices(self):
        modules = self.get_modules()
        if modules is None:
            return

        for k in modules.keys():
            if k not in self.cfg['devices']:
                self.cfg['devices'][k] = {}
                self.cfg['devices'][k]['group_addr'] = str(tools.get_random_uuid())
                self.cfg['devices'][k]['base_addr'] = str(tools.get_random_base_uuid())
                self.cfg['devices'][k]['type'] = modules[k]
                self.cfg['devices'][k].inline_comments['type'] = self.get_comments(modules[k])

        for k in self.cfg['devices']:
            if k not in modules.keys():
                self.cfg['devices'][k]['type'] = 'NoneType'
                self.cfg['devices'][k].inline_comments['type'] = self.get_comments('NoneType')
                self.cfg['devices'].inline_comments[k] = 'This device is removed !!!'

    def get_modules(self):
        logger.info("Fetching module list")
        data = data = API(self.cfg).get_data()
        if data is None:
            logger.warning('Unable to fetch module list')
            return None

        # NAMain is the first devices
        main_module = data['devices'][0]
        result = {main_module['_id']: main_module['type']}
        # other are embedded in modules
        modules = main_module['modules']
        for k in range(0, len(modules)):
            result.update({modules[k]['_id']: modules[k]['type']})
        return result

    def get_comments(self, netatmo_type):
        if netatmo_type == 'NAMain':
            return 'mainIndoor'
        elif netatmo_type == 'NAModule1':
            return 'additionOutdoor'
        elif netatmo_type == 'NAModule2':
            return 'additionWindGauge'
        elif netatmo_type == 'NAModule3':
            return 'additionRainGauge'
        elif netatmo_type == 'NAModule4':
            return 'additionIndoor'
        else:
            return 'invalid netatmo type !!!'



from xaal.lib import tools, Device, helpers

from . import netatmo
import platform
import logging

RATE = 300 # update every 5 min
PACKAGE_NAME = "xaal.netatmo"
logger = logging.getLogger(PACKAGE_NAME)




class GW:
    def __init__(self,engine):
        filename=tools.get_cfg_filename(PACKAGE_NAME)
        self.cfg=tools.load_cfg_file(filename)
        self.eng = engine
        self.modules=[]
        self.setup()
        self.read_weather_stations()
         
    def setup(self):
        # check config file
        config=netatmo.ConfigParser(self.cfg)    
        config.update()

        # load modules from confile file
        self.load_modules_config()

        # GW 
        gw            = Device("gateway.basic")
        gw.address    = tools.get_uuid(self.cfg['config']['addr'])
        gw.vendor_id  = "NETATMO"
        gw.product_id = "Netatmo Weather Station"
        gw.info       = "%s@%s" % (PACKAGE_NAME,platform.node())
        gw.url        = "https://dev.netatmo.com/"
        gw.version    = 0.2
        self.eng.add_device(gw)

        # embedded
        emb = gw.new_attribute('embedded',[]).value
        for mod in self.modules :
            for gw in mod.sensors:
                self.eng.add_device(gw)
                emb.append(gw.address)

        self.eng.add_timer(self.read_weather_stations,RATE)


    def load_modules_config(self):
        """init all modules"""
        assert self.cfg
        devices=self.cfg["devices"]
        for _id,value in devices.items():
            group=tools.get_uuid(value["group_addr"])
            addr=tools.get_uuid(value["base_addr"])
            if value["type"]=="NAMain":
                sensors_type=["thermometer","hygrometer","barometer","co2meter","soundmeter","wifimeter"]
            elif value["type"]=="NAModule1":  # addition outdoor
                sensors_type=["thermometer","hygrometer","radiometer","battery"]
            elif value["type"]=="NAModule4":  #additional indoor
                sensors_type=["thermometer","hygrometer","co2meter","radiometer","battery"]
            elif value["type"]=="NAModule2":  # additon wind gauge
                sensors_type=["windgauge","radiometer","battery"]
            elif value["type"]=="NAModule3":  # addition rain gauge
                sensors_type=["rainmeter","radiometer","battery"]
            else:
                logger.warning("invalid modules!!!")

            module=netatmo.NAModule(addr,group,_id,value['type'])
            module.new_sensors(sensors_type) 
            self.modules.append(module)

    def update_modules(self,macId,attr):
        for m in self.modules :
            if m.macId == macId :
                m.update_sensors(attr)   

    @helpers.spawn
    def read_weather_stations(self):
        netatmo_data = netatmo.API(self.cfg).get_data()
        if netatmo_data is None:
            logger.warning("Unable to read getstationsdata")
            return
        main_indoor = netatmo_data["devices"][0] # dic
        netatmo_modules= main_indoor["modules"] # list

        try:# main indoor
            macId=main_indoor["_id"]
            attr={}
            data = main_indoor["dashboard_data"]
            attr["rssi"] = main_indoor["wifi_status"]            
            attr["temperature"] = data["Temperature"]
            attr["humidity"]    = data["Humidity"]
            attr["pressure"]    = data["Pressure"]
            attr["co2"]         = data["CO2"]
            attr["sound"]       = data["Noise"]

            self.update_modules(macId,attr)
        except Exception as e:
            logger.warning(e)

        try: # update modules
            logger.info("update additional modules")
            for m in netatmo_modules :     
                macId=m["_id"]
                attr={"radio_status": m["rf_status"],
                     "level": m["battery_percent"]}
                data =m.get('dashboard_data',None)
                if not data:
                    logger.warning('Not data from %s' % macId)
                    continue
                try :
                    # addition outdoor
                    if m["type"]=="NAModule1": 
                        attr["temperature"]= data["Temperature"]
                        attr["humidity"]   = data["Humidity"]

                    # additon wind gauge
                    elif m["type"]=="NAModule2":
                        # in API we can also get history of wind, here we don't
                        attr["windstrength"] = data["WindStrength"]
                        attr["windangle"]    = data["WindAngle"]
                        attr["guststrength"] = data["GustStrength"]
                        attr["gustangle"]    = data["GustAngle"]

                    # addition rain gauge
                    elif m["type"]=="NAModule3" : 
                        attr["rain"]=  data["Rain"]

                    #additional indoor 
                    elif m["type"]=="NAModule4" : 
                        attr["temperature"] = data["Temperature"]
                        attr["humidity"]    = data["Humidity"]
                        attr["co2"]         = data["CO2"]
                    else :
                        logger.warning("invalid netatmo module type !!! =>%s" % m["type"])
               
                    self.update_modules(macId,attr)
                except:
                    logger.warning("addition module update failed macId=%s"%macId)

        except Exception as e:
            logger.warning(e)
            logger.warning("Something goes wrong in __readweatherstation")


def setup(engine):
    GW(engine)
    return True

